"""
    Pre-Schema
"""

import strawberry
from graphql.validation import NoSchemaIntrospectionCustomRule
from strawberry.extensions import AddValidationRules, QueryDepthLimiter
from strawberry.tools import merge_types


def Schema(
    query: list = None,
    mutation: list = None,
    extensions: list = None,
    max_depth: int = 4,
    introspection: bool = True,
    **kwargs
) -> strawberry.Schema:
    """Strawberry Schema Wrapper"""

    query = query or []
    mutation = mutation or []
    extensions = extensions or []

    Query = tuple(query)
    Mutation = tuple(mutation)

    # Extension
    Extensions = [
        QueryDepthLimiter(max_depth=max_depth),
    ]
    Extensions.extend(extensions)

    # Introspection
    if not introspection:
        Extensions.append(AddValidationRules([NoSchemaIntrospectionCustomRule]))

    items = {}
    if Query:
        items["query"] = merge_types("Query", Query)
    if Mutation:
        items["mutation"] = merge_types("Mutation", Mutation)

    # Return Value
    return strawberry.Schema(**items, extensions=Extensions, **kwargs)



"""
    Database Manager (with dbcontroller)
    * https://pypi.org/project/dbcontroller/
"""

import pathlib

import motor.motor_asyncio
from databases import Database as DB
from dbcontroller import Mongo, Sql
from pymongo import MongoClient

from ..config import Settings
from ..utils.objects import Singleton


def mongo_url(path):
    """Process Mongo URL"""
    root = pathlib.Path(path)
    return dict(
        database=root.parts[-1:][0].split("?")[0], url="//".join(root.parts[:-1])
    )


class Database(Singleton):
    """Multi Database Manager"""

    def init(self):
        settings = Settings()

        def sql():
            database_url = settings.env.sql or "sqlite:///db.sqlite3"
            if database_url:
                database = DB(database_url)
                return Sql(database)
            return None

        def mongo_motor():
            mongo_env = settings.env.mongo
            if mongo_env:
                active = mongo_url(mongo_env)
                mongo_db = motor.motor_asyncio.AsyncIOMotorClient(
                    active.get("url", "mongodb://localhost:27017")
                )
                return Mongo(mongo_db)(active.get("database"))
            return None

        def mongo():
            mongo_env = settings.env.mongo
            if mongo_env:
                active = mongo_url(mongo_env)
                mongo_client = MongoClient(
                    active.get("url", "mongodb://localhost:27017")
                )
                return mongo_client[active.get("database")]
            return None

        # Start Values
        self.sql = sql()
        self.motor = mongo_motor()
        self.mongo = mongo()



"""Get App(s) Models - Inspire by Django

methods:
    process_models: Collect App Models.
"""

from dbcontroller.manager.mongo import MongoCrud
from dbcontroller.manager.sql import ManagerCrud

from ...utils.objects import get_fields

is_sql = lambda model: isinstance(model, ManagerCrud)
is_mongo = lambda model: isinstance(model, MongoCrud)


def process_models(app_name, app_module):
    """Collect App Models"""
    all_models = {}
    mod_name = app_name.replace(".models", "")
    for item in get_fields(app_module):
        model = getattr(app_module, item)
        is_model = False
        if is_mongo(model):
            model.engine = "mongo"
            is_model = True
        elif is_sql(model):
            model.engine = "sql"
            is_model = True
        if is_model:
            uri_name = f"{mod_name}.{item.lower()}"
            all_models[uri_name] = model
    return all_models



"""
    Class Tools
"""


class ClassPropertyDescriptor:
    """Used to generate the @classproperty"""

    def __init__(self, fget, fset=None):
        self.fget = fget
        self.fset = fset

    def __get__(self, obj, klass=None):
        if klass is None:
            klass = type(obj)
        return self.fget.__get__(obj, klass)()

    def __set__(self, obj, value):
        if not self.fset:
            raise AttributeError("can't set attribute")
        type_ = type(obj)
        return self.fset.__get__(obj, type_)(value)

    def setter(self, func):
        """Set Class-Property"""
        if not isinstance(func, (classmethod, staticmethod)):
            func = classmethod(func)
        self.fset = func
        return self


def classproperty(func):
    """Use to create a Class-Property"""

    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)
    return ClassPropertyDescriptor(func)



"""
    Run FastAPI Server
"""

import json

import click

from ... import Settings
from .shell import shell_commands, shell_print

settings = Settings()


@click.command()
@click.option(
    "-m",
    "--mode",
    type=click.Choice(["development", "staging", "production"]),
    default="development",
    help="Server Mode.",
)
@click.option("-p", "--port", default=8000, help="Port Number.")
@click.option("-w", "--workers", default=1, help="Workers Number.")
def run(mode, port, workers):
    """Start FastApi Server

    \b
    # Examples
        - development : ./manage.py run --port 8001 --mode development
        - staging     : ./manage.py run --port 8001 --mode staging
        - production  : ./manage.py run --mode production
    """

    # SETUP
    return_value = []

    # Get Path(s)
    base_dir = settings.base_dir
    conf_dir = base_dir / "config"

    # Create Path(s)
    conf_dir.mkdir(parents=True, exist_ok=True)

    # Write Mode <File>
    click.secho("")
    with open(conf_dir / "mode.json", "w", encoding="utf-8") as file:
        shell_print(
            f"* Writing to ./config/mode.json ... { json.dumps(dict(mode=mode)) }"
        )
        mode_json = json.dumps(dict(mode=mode), indent=4)
        file.write(mode_json)

    # PY-Lint
    if mode == "development":
        return_value.append("""python scripts/pylint.py > logs/pylint.log 2>&1 &""")

    # FastAPI-Server
    match mode:
        case "development":
            run_server = f"python -m uvicorn main:app --reload --port { port }"
        case "staging":
            run_server = f"python -m uvicorn main:app --reload --port { port }"
        case "production":
            run_server = f"python -m gunicorn main:app --workers { workers } --worker-class uvicorn.workers.UvicornWorker --bind 0.0.0.0:{ port }"
                    
    return_value.append(run_server)

    # Message
    shell_print(f"* Starting FastApi Server... (Mode: { mode })\n")

    # RETURN-VALUE
    shell_commands(return_value)
